"""File contains definition of CLI program config data model."""

from dataclasses import dataclass
from typing import List, Optional

from foxai.context_manager import ExplainerWithParams


@dataclass
class MethodDataModel:
    """Method data model for CLI program."""

    explainer_with_params: ExplainerWithParams
    artifact_name: Optional[str] = None


@dataclass
class ConfigDataModel:
    """Config data model for CLI program."""

    method_config_list: List[MethodDataModel]



# pylint: disable = missing-module-docstring
from foxai.explainer.computer_vision.algorithm.conductance import (
    LayerConductanceCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.deconv import DeconvolutionCVExplainer
from foxai.explainer.computer_vision.algorithm.deeplift import (
    DeepLIFTCVExplainer,
    LayerDeepLIFTCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.deeplift_shap import (
    DeepLIFTSHAPCVExplainer,
    LayerDeepLIFTSHAPCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.gradcam import (
    GuidedGradCAMCVExplainer,
    LayerGradCAMCVExplainer,
    LayerGradCAMObjectDetectionExplainer,
)
from foxai.explainer.computer_vision.algorithm.gradient_shap import (
    GradientSHAPCVExplainer,
    LayerGradientSHAPCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.guided_backprop import (
    GuidedBackpropCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.input_x_gradient import (
    InputXGradientCVExplainer,
    LayerInputXGradientCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.integrated_gradients import (
    IntegratedGradientsCVExplainer,
    LayerIntegratedGradientsCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.lrp import (
    LayerLRPCVExplainer,
    LRPCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.noise_tunnel import (
    LayerNoiseTunnelCVExplainer,
    NoiseTunnelCVExplainer,
)
from foxai.explainer.computer_vision.algorithm.occlusion import OcclusionCVExplainer
from foxai.explainer.computer_vision.algorithm.saliency import SaliencyCVExplainer



"""File with Saliency algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/saliency.py.
"""

from typing import Any

import torch
from captum._utils.typing import TargetType
from captum.attr import Saliency

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer


class SaliencyCVExplainer(Explainer):
    """Saliency algorithm explainer."""

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        abs_value: bool = True,
        additional_forward_args: Any = None,
        **kwargs,  # pylint: disable = (unused-argument)
    ) -> torch.Tensor:
        """Generate model's attributes with Saliency algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which saliency
                is computed. If forward_func takes a single tensor
                as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            abs_value: Returns absolute value of gradients if set
                to True, otherwise returns the (signed) gradients if
                False.
                Default: True
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None

        Returns:
            The gradients with respect to each input feature.
            Attributions will always be
            the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        saliency = Saliency(forward_func=model)

        attributions = saliency.attribute(
            input_data,
            target=pred_label_idx,
            abs=abs_value,
            additional_forward_args=additional_forward_args,
        )
        validate_result(attributions=attributions)
        return attributions



"""File with Guided Backpropagation algorithm explainer class.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/guided_backprop_deconvnet.py.
"""

from abc import abstractmethod
from typing import Any

import torch
from captum._utils.typing import TargetType
from captum.attr import GuidedBackprop

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import modify_modules


class BaseGuidedBackpropCVExplainer(Explainer):
    """Base Guided Backpropagation algorithm explainer."""

    @abstractmethod
    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> GuidedBackprop:
        """Create explainer object.

        Returns:
            Explainer object.
        """

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        additional_forward_args: Any = None,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Guided Backpropagation algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which
                attributions are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
                If forward_func takes multiple tensors as input, a tuple
                of the input tensors should be provided. It is assumed
                that for all given input tensors, dimension 0 corresponds
                to the number of examples (aka batch size), and if
                multiple input tensors are provided, the examples must
                be aligned appropriately.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a tuple
                containing multiple additional arguments including tensors
                or any arbitrary python types. These arguments are provided to
                forward_func in order, following the arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None

        Returns:
            The guided backprop gradients with respect to each
            input feature. Attributions will always
            be the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        guided_backprop = self.create_explainer(model=model)
        attributions = guided_backprop.attribute(
            input_data,
            target=pred_label_idx,
            additional_forward_args=additional_forward_args,
        )
        validate_result(attributions=attributions)
        return attributions


class GuidedBackpropCVExplainer(BaseGuidedBackpropCVExplainer):
    """Guided Backpropagation algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> GuidedBackprop:
        """Create explainer object.

        Returns:
            Explainer object.
        """

        return GuidedBackprop(model=modify_modules(model))



"""File with Noise Tunnel algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/noise_tunnel.py.
"""

from abc import abstractmethod
from typing import Optional, Tuple, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import IntegratedGradients, LayerIntegratedGradients, NoiseTunnel

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import get_last_conv_model_layer


class BaseNoiseTunnelCVExplainer(Explainer):
    """Base Noise Tunnel algorithm explainer."""

    @abstractmethod
    def create_explainer(self, model: torch.nn.Module, **kwargs) -> NoiseTunnel:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        nt_type: str = "smoothgrad",
        nt_samples: int = 5,
        nt_samples_batch_size: Optional[int] = None,
        stdevs: Union[float, Tuple[float, ...]] = 1.0,
        draw_baseline_from_distrib: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Noise Tunnel algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which integrated
                gradients are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            nt_type: Smoothing type of the attributions.
                `smoothgrad`, `smoothgrad_sq` or `vargrad`
                Default: `smoothgrad` if `type` is not provided.
            nt_samples: The number of randomly generated examples
                per sample in the input batch. Random examples are
                generated by adding gaussian random noise to each sample.
                Default: `5` if `nt_samples` is not provided.
            nt_samples_batch_size: The number of the `nt_samples`
                that will be processed together. With the help
                of this parameter we can avoid out of memory situation and
                reduce the number of randomly generated examples per sample
                in each batch.
                Default: None if `nt_samples_batch_size` is not provided. In
                this case all `nt_samples` will be processed together.
            stdevs: The standard deviation
                of gaussian noise with zero mean that is added to each
                input in the batch. If `stdevs` is a single float value
                then that same value is used for all inputs. If it is
                a tuple, then it must have the same length as the inputs
                tuple. In this case, each stdev value in the stdevs tuple
                corresponds to the input with the same index in the inputs
                tuple.
                Default: `1.0` if `stdevs` is not provided.
            draw_baseline_from_distrib: Indicates whether to
                randomly draw baseline samples from the `baselines`
                distribution provided as an input tensor.
                Default: False

        Returns:
            Attribution with respect to each input feature. attributions
            will always be the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)

        noise_tunnel = self.create_explainer(model=model, layer=layer)

        attributions = noise_tunnel.attribute(
            inputs=input_data,
            nt_type=nt_type,
            nt_samples=nt_samples,
            nt_samples_batch_size=nt_samples_batch_size,
            stdevs=stdevs,
            draw_baseline_from_distrib=draw_baseline_from_distrib,
            target=pred_label_idx,
        )
        validate_result(attributions=attributions)
        return attributions


class NoiseTunnelCVExplainer(BaseNoiseTunnelCVExplainer):
    """Noise Tunnel algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> NoiseTunnel:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """
        integrated_gradients = IntegratedGradients(forward_func=model)
        return NoiseTunnel(integrated_gradients)


class LayerNoiseTunnelCVExplainer(BaseNoiseTunnelCVExplainer):
    """Layer Noise Tunnel algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: Optional[torch.nn.Module] = None,
        **kwargs,
    ) -> NoiseTunnel:
        """Create explainer object.

        Uses parameter `layer` from `kwargs`. If not provided function will call
        `get_last_conv_model_layer` function to obtain last `torch.nn.Conv2d` layer
        from provided model.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.
                Default: None

        Returns:
            Explainer object.

        Raises:
            ValueError: if model does not contain conv layers.
        """
        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        integrated_gradients = LayerIntegratedGradients(forward_func=model, layer=layer)
        return NoiseTunnel(integrated_gradients)



"""File with LRP algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/lrp.py
and https://github.com/pytorch/captum/blob/master/captum/attr/_core/layer/layer_lrp.py.
"""

from abc import abstractmethod
from typing import Any, Optional, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import LRP, LayerLRP
from captum.attr._utils.lrp_rules import EpsilonRule, GammaRule

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import (
    get_last_conv_model_layer,
    modify_modules,
)


class BaseLRPCVExplainer(Explainer):
    """Base LRP algorithm explainer."""

    @abstractmethod
    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> Union[LRP, LayerLRP]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with LRP algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which relevance is
                propagated. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a tuple
                containing multiple additional arguments including tensors
                or any arbitrary python types. These arguments are provided to
                forward_func in order, following the arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output. If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer input, otherwise it will be computed with respect
                to layer output.
            verbose: Indicates whether information on application
                of rules is printed during propagation.
                Default: False

        Returns:
            Features matrix.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)

        lrp = self.create_explainer(model=model, layer=layer)

        if isinstance(lrp, LRP):
            attributions = lrp.attribute(
                input_data,
                target=pred_label_idx,
                additional_forward_args=additional_forward_args,
                return_convergence_delta=False,
                verbose=verbose,
            )
        else:
            attributions = lrp.attribute(
                input_data,
                target=pred_label_idx,
                additional_forward_args=additional_forward_args,
                return_convergence_delta=False,
                attribute_to_layer_input=attribute_to_layer_input,
                verbose=verbose,
            )
        validate_result(attributions=attributions)
        return attributions

    def add_rules(self, model: torch.nn.Module) -> torch.nn.Module:
        """Add rules for the LRP explainer,
        according to https://arxiv.org/pdf/1910.09840.pdf.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Modified DNN object.
        """
        layers_number: int = len(list(model.modules()))
        for idx_layer, module in enumerate(model.modules()):
            if idx_layer <= layers_number // 2:
                setattr(module, "rule", GammaRule())
            elif idx_layer != (layers_number - 1):
                setattr(module, "rule", EpsilonRule())
            else:
                setattr(module, "rule", EpsilonRule(epsilon=0))  # LRP-0

        return model


class LRPCVExplainer(BaseLRPCVExplainer):
    """LRP algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> Union[LRP, LayerLRP]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """
        model = self.add_rules(modify_modules(model))

        return LRP(model=model)


class LayerLRPCVExplainer(BaseLRPCVExplainer):
    """Layer LRP algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: Optional[torch.nn.Module] = None,
        **kwargs,
    ) -> Union[LRP, LayerLRP]:
        """Create explainer object.

        Uses parameter `layer` from `kwargs`. If not provided function will call
        `get_last_conv_model_layer` function to obtain last `torch.nn.Conv2d` layer
        from provided model.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.
                Default: None

        Returns:
            Explainer object.

        Raises:
            ValueError: if model does not contain conv layers.
        """
        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        model = self.add_rules(modify_modules(model))

        return LayerLRP(model=model, layer=layer)



"""File contains CLI application for updating W&B experiment artifacts."""

import os
from typing import Any, Dict, List, cast

import hydra
import pandas as pd
import pytorch_lightning as pl
import torch
import wandb
from omegaconf import DictConfig
from torchvision.io import ImageReadMode, read_image

from foxai.cli.config_model import ConfigDataModel, MethodDataModel
from foxai.context_manager import (
    CVClassificationExplainers,
    ExplainerWithParams,
    FoXaiExplainer,
)


def download_upload_metadata(api_run: wandb.apis.public.Run) -> Dict[str, Any]:
    """Fetch uploaded data samples metadata from logs history.

    Args:
        run_api: WandB Api object.

    Returns:
        Metadata of upload data sample entry.
    """
    # get uploaded sample data from experiment run
    history: pd.DataFrame = api_run.history()

    # get log entry with saving all sample data images from log history
    upload_data_metadata = history["validation_data"][
        history["validation_data"].notnull()
    ].iloc[0]

    return upload_data_metadata


def get_labels(upload_data_metadata: Dict[str, Any]) -> List[int]:
    """Extract and convert ground truth labels from image captions.

    Args:
        upload_data_metadata: Metadata of upload data sample entry.

    Returns:
        List of labels.
    """
    # get ground truth labels from captions
    captions: List[str] = upload_data_metadata["captions"]
    labels: List[int] = [int(val.replace("label: ", "")) for val in captions]
    return labels


def fetch_model_checkpoints(
    api: wandb.Api, model_artifact_root: str, project: str
) -> List[str]:
    """Fetch all model checkpoints and return local paths to them.

    Args:
        api: WandB Api object.
        model_artifact_root: Path to local directory where model checkpoints
            will be saved.
        project: Name of a project.

    Returns:
        List of paths to saved downloaded model checkpoints.
    """
    # https://stackoverflow.com/questions/68952727/wandb-get-a-list-of-all-artifact-collections-and-all-aliases-of-those-artifacts
    # list all artifacts (saved model checkpoints)
    collections = list(
        api.artifact_type(type_name="model", project=project).collections()
    )
    model_paths: List[str] = []
    for coll in collections:
        for artifact in coll.versions():
            model_version_root = os.path.join(
                model_artifact_root,
                artifact._artifact_name,  # pylint: disable = (protected-access)
            )
            datadir = artifact.download(root=model_version_root)
            model_paths.append(os.path.join(datadir, "model.ckpt"))

    return model_paths


def load_images_to_tensors(
    filenames: List[str], artifacts_path: str
) -> List[torch.Tensor]:
    """Load images from local directory.

    Args:
        filenames: List of filepaths to images from metadata.
        artifacts_path: Root directory of downloaded images.

    Returns:
        List of tensors representing images.
    """
    # create dataset from downloaded data samples
    image_list: List[torch.Tensor] = []
    for filename in filenames:
        path: str = os.path.join(artifacts_path, filename)
        image: torch.Tensor = read_image(path=path, mode=ImageReadMode.GRAY)
        image_list.append(image)

    return image_list


def load_config(config: DictConfig) -> ConfigDataModel:
    """Load and validate config with explainers to apply.

    Args:
        config: DictConfig object loaded from yaml config file.

    Returns:
        ConfigDataModel object.
    """

    method_config_list: List[MethodDataModel] = []
    for entry in config["explainers"]:
        explainer_config: Dict[str, Any] = entry["explainer_with_params"]
        explainer: ExplainerWithParams = ExplainerWithParams(
            explainer_name=CVClassificationExplainers[
                explainer_config["explainer_name"]
            ],
            kwargs=explainer_config.get("kwargs", {}),
        )
        method_config_list.append(
            MethodDataModel(
                explainer_with_params=explainer,
                artifact_name=entry.get("artifact_name", None),
            )
        )

    config_data = ConfigDataModel(method_config_list=method_config_list)
    return config_data


@hydra.main()
def main(cfg: DictConfig) -> None:  # pylint: disable = (too-many-locals)
    """Entry point for CLI application."""

    model_class: pl.LightningModule = hydra.utils.instantiate(cfg.classifier)
    wandb.login()

    config: ConfigDataModel = load_config(cfg)

    # resume experiment run that has to be updated
    run = wandb.init(
        project=cfg.experiment,
        id=cfg.run_id,
        resume="allow",
    )
    if run is None:
        return

    artifacts_path: str = run.settings.files_dir
    image_artifacts_path: str = os.path.join(artifacts_path, "media", "images")
    if not os.path.exists(image_artifacts_path):
        os.makedirs(image_artifacts_path)

    api = wandb.Api()
    run_api: wandb.apis.public.Run = api.run(
        f"{cfg.username}/{cfg.experiment}/{cfg.run_id}"
    )

    upload_data_metadata = download_upload_metadata(api_run=run_api)

    filenames: List[str] = upload_data_metadata["filenames"]
    # download all data sample images
    for filename in filenames:
        run_api.file(filename).download(root=artifacts_path)

    labels: List[int] = get_labels(upload_data_metadata=upload_data_metadata)

    model_artifact_root: str = os.path.join(run.settings.tmp_dir, "models")
    model_paths: List[str] = fetch_model_checkpoints(
        api=api,
        model_artifact_root=model_artifact_root,
        project=cfg.experiment,
    )

    image_list: List[torch.Tensor] = load_images_to_tensors(
        filenames=filenames,
        artifacts_path=artifacts_path,
    )

    # sort paths from the earliest model to the latest
    # wand.log has internal counter and to have order in explanations we have to sort
    # them ascending by version
    sorted_paths = sorted(
        model_paths, key=lambda x: int(x.split("/")[-2].split(":v")[-1])
    )
    # filter out artifacts from different runs
    sorted_paths = [
        val for val in sorted_paths if cfg.run_id in val.split("/")[-2].split(":")[0]
    ]

    for explainer_config in config.method_config_list:
        artifact_name: str
        if explainer_config.artifact_name is None:
            artifact_name = explainer_config.explainer_with_params.explainer_name.name
        else:
            artifact_name = explainer_config.artifact_name

        for path in sorted_paths:
            model: torch.nn.Module = model_class.load_from_checkpoint(path)
            device: torch.device = cast(torch.device, model.device)

            explanations: List[wandb.Image] = []
            for input_data, label in zip(image_list, labels):
                with FoXaiExplainer(
                    model=model,
                    explainers=[explainer_config.explainer_with_params],
                    target=label,
                ) as xai_model:
                    input_data = input_data.float()
                    _, attributes_dict = xai_model(input_data.to(device))
                    explainer_name: str = (
                        explainer_config.explainer_with_params.explainer_name.name
                    )
                    attributes: torch.Tensor = attributes_dict[explainer_name]
                    explanations.append(
                        wandb.Image(attributes, caption=f"label: {label}")
                    )

            if explanations:
                wandb.log({artifact_name: explanations})



"""
File contains functions to handle images, coordinates and calcualte IoU.

Based on code: https://github.com/pooya-mohammadi/yolov5-gradcam.
"""

from typing import Tuple

import cv2
import numpy as np


def resize_image(
    image: np.ndarray,
    new_shape: Tuple[int, int] = (640, 640),
    change_original_ratio: bool = False,
    scaleup: bool = True,
    interpolation: int = cv2.INTER_LINEAR,
) -> np.ndarray:
    """Resize image to given shape.

    Args:
        image: Image to resize.
        new_shape: Desired shape of image. Defaults to (640, 640).
        change_original_ratio: If resized image should have different height to
            width ratio than original image. Defaults to False.
        scaleup: If scale up image. Defaults to True.
        interpolation: OpenCV interpolation method. Defaults to `cv2.INTER_LINEAR`.

    Returns:
        Resized image.
    """
    # Resize and pad image while meeting stride-multiple constraints
    # image has shape (H x W x C)
    shape = image.shape[:2]  # current shape [height, width]

    # Scale ratio (new / old)
    # get minimum of width and height ratios
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # do not change the width/height ratio of the image, assign the same ratio
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    if change_original_ratio:  # stretch
        new_unpad = (new_shape[1], new_shape[0])

    # if current shape is different than desired shape call resize function
    if shape[::-1] != new_unpad:  # resize
        image = cv2.resize(image, new_unpad, interpolation=interpolation)

    return image



"""Callback for Weights and Biases."""
from collections import defaultdict
from typing import Dict, Generator, List, Optional, Tuple, cast

import matplotlib
import numpy as np
import pytorch_lightning as pl
import torch
import wandb
from pytorch_lightning.loggers import WandbLogger
from torch.utils.data import DataLoader

from foxai.array_utils import convert_standardized_float_to_uint8, standardize_array
from foxai.context_manager import ExplainerWithParams, FoXaiExplainer
from foxai.visualizer import mean_channels_visualization

AttributeMapType = Dict[str, List[np.ndarray]]
CaptionMapType = Dict[str, List[str]]
FigureMapType = Dict[str, List[matplotlib.pyplot.Figure]]


class WandBCallback(pl.callbacks.Callback):
    """Library callback for Weights and Biases."""

    def __init__(  # pylint: disable = (too-many-arguments)
        self,
        wandb_logger: WandbLogger,
        explainers: List[ExplainerWithParams],
        idx_to_label: Dict[int, str],
        max_artifacts: int = 3,
    ):
        """Initialize Callback class.

        Args:
            wandb_logger: Pytorch-lightning wandb logger.
            idx_to_label: Index to label mapping.
            explainers: List of explainer algorithms of type ExplainerWithParams.
            idx_to_label: Dictionary with mapping from model index to label.
            max_artifacts: Number of maximum number of artifacts to be logged.
                Defaults to 3.
        """
        super().__init__()
        self.explainers = explainers
        self.wandb_logger = wandb_logger
        self.idx_to_label = idx_to_label
        self.max_artifacts = max_artifacts

    def _save_idx_mapping(self) -> None:
        """Saving index to label mapping to experiment logs directory."""
        self.wandb_logger.log_table(
            key="idx2label",
            columns=["index", "label"],
            data=[[key, val] for key, val in self.idx_to_label.items()],
        )

    def iterate_dataloader(
        self, dataloader_list: List[DataLoader], max_items: int
    ) -> Generator[Tuple[torch.Tensor, torch.Tensor], None, None]:
        """Iterate over dataloader list with constraint on max items returned.

        Args:
            dataloader: Trainer dataloader.
            max_items: Max items to return.

        Yields:
            Tuple containing training sample and corresponding label.
        """
        index: int = 0
        dataloader: DataLoader
        item: torch.Tensor
        target_label: torch.Tensor
        for dataloader in dataloader_list:
            for batch in dataloader:
                for item, target_label in zip(*batch):
                    if index >= max_items:
                        break

                    index += 1
                    yield item, target_label

    def explain(  # pylint: disable = (too-many-arguments)
        self,
        model: pl.LightningModule,
        item: torch.Tensor,
        target_label: torch.Tensor,
        attributes_dict: AttributeMapType,
        caption_dict: CaptionMapType,
        figures_dict: FigureMapType,
    ) -> Tuple[AttributeMapType, CaptionMapType, FigureMapType,]:
        """Calculate explainer attributes, creates captions and figures.

        Args:
            model: Model to explain.
            item: Input data sample tensor.
            target_label: Sample label.
            attributes_dict: List of attributes for every explainer and sample.
            caption_dict: List of captions for every explainer and sample.
            figures_dict: List of figures for every explainer and sample.

        Returns:
            Tuple of maps containing attributes, captions and figures for
            every explainer and sample.
        """
        with FoXaiExplainer(
            model=model,
            explainers=self.explainers,
            target=int(target_label.item()),
        ) as xai_model:
            _, attributes = xai_model(item.to(cast(torch.device, model.device)))

        for explainer in self.explainers:
            explainer_name: str = explainer.explainer_name.name
            explainer_attributes: torch.Tensor = attributes[explainer_name]
            caption_dict[explainer_name].append(f"label: {target_label}")
            figure = mean_channels_visualization(
                attributions=explainer_attributes,
                transformed_img=item,
            )
            figures_dict[explainer_name].append(figure)
            standardized_attr = standardize_array(
                explainer_attributes.detach().cpu().numpy().astype(float)
            )

            attributes_dict[explainer_name].append(
                convert_standardized_float_to_uint8(standardized_attr),
            )

        return attributes_dict, caption_dict, figures_dict

    def on_train_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,  # pylint: disable = (unused-argument)
    ) -> None:
        """Save index to labels mapping and validation samples to experiment
        at `fit`.

        Args:
            trainer: Trainer object.
            pl_module: Model to explain.
        """
        if trainer.val_dataloaders is None:
            return

        self._save_idx_mapping()

        image_matrix: Optional[torch.Tensor] = None
        image_labels: List[str] = []

        for item, target_label in self.iterate_dataloader(
            dataloader_list=trainer.val_dataloaders, max_items=self.max_artifacts
        ):
            if image_matrix is None:
                image_matrix = item
            else:
                image_matrix = torch.cat(  # pylint: disable = (no-member)
                    [image_matrix, item]
                )

            image_labels.append(f"label: {target_label.item()}")

        if image_matrix is None:
            return

        list_of_images: List[torch.Tensor] = list(torch.split(image_matrix, 1))
        self.wandb_logger.log_image(
            key="validation_data",
            images=list_of_images[: min(len(list_of_images), self.max_artifacts)],
            caption=image_labels[: min(len(image_labels), self.max_artifacts)],
        )

    def on_validation_epoch_end(  # pylint: disable = (too-many-arguments, too-many-locals)
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
    ) -> None:
        """Export model's state dict in log directory on validation epoch end.

        Args:
            trainer: Trainer object.
            pl_module: Model to explain.
        """
        if trainer.val_dataloaders is None:
            return

        attributes_dict: AttributeMapType = defaultdict(list)
        caption_dict: CaptionMapType = defaultdict(list)
        figures_dict: FigureMapType = defaultdict(list)

        for item, target_label in self.iterate_dataloader(
            dataloader_list=trainer.val_dataloaders,
            max_items=self.max_artifacts,
        ):
            attributes_dict, caption_dict, figures_dict = self.explain(
                model=pl_module,
                item=item,
                target_label=target_label,
                attributes_dict=attributes_dict,
                caption_dict=caption_dict,
                figures_dict=figures_dict,
            )

        self.log_explanations(
            attributes_dict=attributes_dict,
            caption_dict=caption_dict,
            figures_dict=figures_dict,
        )

    def log_explanations(
        self,
        attributes_dict: AttributeMapType,
        caption_dict: CaptionMapType,
        figures_dict: FigureMapType,
    ) -> None:
        """Log explanation artifacts to W&B experiment.

        Args:
            attributes_dict: Numpy array attributes for every sample and every explainer.
            caption_dict: Caption for every sample and every explainer.
            figures_dict: Figure with attributes for every sample and every explainer.
        """
        # upload artifacts to the wandb experiment
        for explainer in self.explainers:
            explainer_name: str = explainer.explainer_name.name
            self.wandb_logger.log_image(
                key=f"{explainer_name}",
                images=attributes_dict[explainer_name],
                caption=caption_dict[explainer_name],
            )

            # matplotlib Figures can not be directly logged via WandbLogger
            # we have to use native Run object from wandb which is more powerfull
            wandb_image_list: List[wandb.Image] = []
            for figure in figures_dict[explainer_name]:
                wandb_image_list.append(wandb.Image(figure))

            self.wandb_logger.experiment.log(
                {f"{explainer_name}_explanations": wandb_image_list}
            )



"""
Run xai alongside with inference.

Example:
    with FoXaiExplainer(
        model=classifier,
        explainers=[
            ExplainerWithParams(
                explainer_name=Explainers.CV_GRADIENT_SHAP_EXPLAINER,
                n_samples=100,
                stdevs=0.0005,
            ),
        ],
        target=pred_label_idx,
    ) as xai_model:
        output, xai_explanations = xai_model(img_tensor)
"""
import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, Tuple, Union, cast

import torch

from foxai import explainer
from foxai.explainer import (
    DeconvolutionCVExplainer,
    DeepLIFTCVExplainer,
    DeepLIFTSHAPCVExplainer,
    GradientSHAPCVExplainer,
    GuidedBackpropCVExplainer,
    GuidedGradCAMCVExplainer,
    InputXGradientCVExplainer,
    IntegratedGradientsCVExplainer,
    LayerConductanceCVExplainer,
    LayerDeepLIFTCVExplainer,
    LayerDeepLIFTSHAPCVExplainer,
    LayerGradCAMCVExplainer,
    LayerGradCAMObjectDetectionExplainer,
    LayerGradientSHAPCVExplainer,
    LayerInputXGradientCVExplainer,
    LayerIntegratedGradientsCVExplainer,
    LayerLRPCVExplainer,
    LayerNoiseTunnelCVExplainer,
    LRPCVExplainer,
    NoiseTunnelCVExplainer,
    OcclusionCVExplainer,
    SaliencyCVExplainer,
)
from foxai.explainer.base_explainer import CVExplainerT
from foxai.logger import create_logger

_LOGGER: Optional[logging.Logger] = None


def log() -> logging.Logger:
    """Get or create logger."""
    # pylint: disable = global-statement
    global _LOGGER
    if _LOGGER is None:
        _LOGGER = create_logger(__name__)
    return _LOGGER


class CVClassificationExplainers(Enum):
    """Enum of supported computer vision classification explainers types."""

    CV_OCCLUSION_EXPLAINER: str = OcclusionCVExplainer.__name__
    CV_INTEGRATED_GRADIENTS_EXPLAINER: str = IntegratedGradientsCVExplainer.__name__
    CV_NOISE_TUNNEL_EXPLAINER: str = NoiseTunnelCVExplainer.__name__
    CV_GRADIENT_SHAP_EXPLAINER: str = GradientSHAPCVExplainer.__name__
    CV_LRP_EXPLAINER: str = LRPCVExplainer.__name__
    CV_GUIDEDGRADCAM_EXPLAINER: str = GuidedGradCAMCVExplainer.__name__
    CV_LAYER_INTEGRATED_GRADIENTS_EXPLAINER: str = (
        LayerIntegratedGradientsCVExplainer.__name__
    )
    CV_LAYER_NOISE_TUNNEL_EXPLAINER: str = LayerNoiseTunnelCVExplainer.__name__
    CV_LAYER_GRADIENT_SHAP_EXPLAINER: str = LayerGradientSHAPCVExplainer.__name__
    CV_LAYER_LRP_EXPLAINER: str = LayerLRPCVExplainer.__name__
    CV_LAYER_GRADCAM_EXPLAINER: str = LayerGradCAMCVExplainer.__name__
    CV_INPUT_X_GRADIENT_EXPLAINER: str = InputXGradientCVExplainer.__name__
    CV_LAYER_INPUT_X_GRADIENT_EXPLAINER: str = LayerInputXGradientCVExplainer.__name__
    CV_DEEPLIFT_EXPLAINER: str = DeepLIFTCVExplainer.__name__
    CV_LAYER_DEEPLIFT_EXPLAINER: str = LayerDeepLIFTCVExplainer.__name__
    CV_DEEPLIFT_SHAP_EXPLAINER: str = DeepLIFTSHAPCVExplainer.__name__
    CV_LAYER_DEEPLIFT_SHAP_EXPLAINER: str = LayerDeepLIFTSHAPCVExplainer.__name__
    CV_DECONVOLUTION_EXPLAINER: str = DeconvolutionCVExplainer.__name__
    CV_LAYER_CONDUCTANCE_EXPLAINER: str = LayerConductanceCVExplainer.__name__
    CV_SALIENCY_EXPLAINER: str = SaliencyCVExplainer.__name__
    CV_GUIDED_BACKPOPAGATION_EXPLAINER: str = GuidedBackpropCVExplainer.__name__


class CVObjectDetectionExplainers(Enum):
    """Enum of supported computer vision object detection explainers types."""

    CV_LAYER_GRADCAM_OBJECT_DETECTION_EXPLAINER: str = (
        LayerGradCAMObjectDetectionExplainer.__name__
    )


@dataclass
class ExplainerWithParams:
    """Holder for explainer name (class name) and it's params"""

    explainer_name: Union[CVClassificationExplainers, CVObjectDetectionExplainers]
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def __init__(
        self,
        explainer_name: Union[CVClassificationExplainers, CVObjectDetectionExplainers],
        **kwargs
    ) -> None:
        self.explainer_name = explainer_name
        if kwargs:
            self.kwargs = kwargs
        else:
            self.kwargs = {}


@dataclass
class ExplainerClassWithParams(Generic[CVExplainerT]):
    """Holder for explainer class and it's params"""

    explainer_class: CVExplainerT
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def __init__(self, explainer_class: CVExplainerT, **kwargs) -> None:
        self.explainer_class = explainer_class
        if kwargs:
            self.kwargs = kwargs
        else:
            self.kwargs = {}


class FoXaiExplainer(Generic[CVExplainerT]):
    """Context menager for FoXAI explanation.

    Example:
        with FoXaiExplainer(
            model=classifier,
            explainers=[
                ExplainerWithParams(
                    explainer_name=Explainers.CV_GRADIENT_SHAP_EXPLAINER,
                    n_samples=100,
                    stdevs=0.0005,
                ),
            ],
            target=pred_label_idx,
        ) as xai_model:
            output, xai_explanations = xai_model(img_tensor)

    Raises:
        ValueError: if no explainer provided
    """

    def __init__(
        self,
        model: torch.nn.Module,
        explainers: List[ExplainerWithParams],
        target: int = 0,
    ) -> None:
        """
        Args:
            model: the torch model to exavluate with CV explainer
            explainers: explainers names list, to use for model evaluation.
            target: predicted target index. For which class to generate xai.
        """

        if not explainers:
            raise ValueError("At leas one explainer should be defined.")

        self.model: torch.nn.Module = model
        self.prev_model_training_state: bool = self.model.training

        self.explainer_map: Dict[str, ExplainerClassWithParams] = {
            explainer_with_params.explainer_name.name: ExplainerClassWithParams(
                explainer_class=getattr(
                    explainer, explainer_with_params.explainer_name.value
                )(),
                **explainer_with_params.kwargs,
            )
            for explainer_with_params in explainers
        }

        self.target: int = target

    def __enter__(self) -> "FoXaiExplainer":
        """Verify if model is in eval() mode.

        Raises:
            ValueError: if the model is in training mode.

        Returns:
            the foxai class instance.
        """
        self.prev_torch_grad = torch.is_grad_enabled()
        if not self.prev_torch_grad:
            log_msg: str = (
                "Torch model explainer can be called only with enabled "
                + "gradients, as it depends on gradients computations. The model is going "
                + "to be toggled to gradients enabled. For the "
                + "model prediction, the gradient is temporary turned off."
            )
            log().warning(log_msg)
            torch.set_grad_enabled(True)

        if self.model.training:
            self.model.eval()
            log().warning(
                "The model should be in the eval model. Toggling it to eval mode right now."
            )

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """If the torch was not recording gradient, before entering in the
        context manager mode, switch it back to no gradient recording mode.

        If the torch was recording gradient before entering in the context
        manager modes, nothings changes.

        Setup model to previous state: `eval` or `training` to match initial
        state.
        """
        torch.set_grad_enabled(self.prev_torch_grad)
        self.model.train(self.prev_model_training_state)

    def __call__(self, *args, **kwargs) -> Tuple[Any, Dict[str, torch.Tensor]]:
        """Run model prediction and explain the model with given explainers.

        Explainers and model are defined as the class parameter.

        Args:
            list of arguments for the torch.nn.Module forward method.

        Returns:
            the model output and explanations for each requested explainer.
        """

        with torch.no_grad():
            model_output: Any = self.model(*args, **kwargs)

        if len(args) != 1:
            # TODO: add support in explainer for multiple input models
            raise NotImplementedError(
                "calculate_features() functions "
                + "in explainers does not support multiple inputs to the model."
            )
        input_tensor: torch.Tensor = cast(torch.Tensor, args)[0]
        # cashe tensor requires grad state
        prev_requires_grad: bool = input_tensor.requires_grad
        # turn on requires grad for the input tensor
        input_tensor.requires_grad = True

        explanations: Dict[str, torch.Tensor] = {}
        for explainer_name in self.explainer_map:
            # zero the previous gradient for the model
            self.model.zero_grad()
            # run explainer
            explainer_kwargs: Dict[str, Any] = self.explainer_map[explainer_name].kwargs
            explainer_class: CVExplainerT = self.explainer_map[
                explainer_name
            ].explainer_class

            explanations[explainer_name] = (
                explainer_class.calculate_features(
                    model=self.model,
                    input_data=input_tensor,
                    pred_label_idx=self.target,
                    **explainer_kwargs,
                )
                .detach()
                .cpu()
            )
            input_tensor.grad = None

        # restore tensor requires grad state
        input_tensor.requires_grad = prev_requires_grad

        return model_output, explanations



"""File with Conductance algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/layer/layer_conductance.py.
"""

from typing import Any, Optional, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import LayerConductance

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import get_last_conv_model_layer


class LayerConductanceCVExplainer(Explainer):
    """Layer Conductance algorithm explainer."""

    # pylint: disable = unused-argument
    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: torch.nn.Module,
    ) -> LayerConductance:
        """Create explainer object.

        model: The forward function of the model or any
            modification of it.
        layer: Layer for which attributions are computed.
            Output size of attribute matches this layer's input or
            output dimensions, depending on whether we attribute to
            the inputs or outputs of the layer, corresponding to
            attribution of each neuron in the input or output of
            this layer.

        Returns:
            Explainer object.
        """
        conductance = LayerConductance(forward_func=model, layer=layer)

        return conductance

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        baselines: Union[None, int, float, torch.Tensor] = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        internal_batch_size: Union[None, int] = None,
        attribute_to_layer_input: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Layer Conductance algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which layer
                conductance is computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.
                Default: None
            baselines:
                Baselines define the starting point from which integral
                is computed and can be provided as:

                - a single tensor, if inputs is a single tensor, with
                    exactly the same dimensions as inputs or the first
                    dimension is one and the remaining dimensions match
                    with inputs.

                - a single scalar, if inputs is a single tensor, which will
                    be broadcasted for each input value in input tensor.

                In the cases when `baselines` is not provided, we internally
                use zero scalar corresponding to each input tensor.
                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                For a tensor, the first dimension of the tensor must
                correspond to the number of examples. It will be repeated
                for each of `n_steps` along the integrated path.
                For all other types, the given argument is used for
                all forward evaluations.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            n_steps: The number of steps used by the approximation
                method. Default: 50.
            method: Method for approximating the integral,
                one of `riemann_right`, `riemann_left`, `riemann_middle`,
                `riemann_trapezoid` or `gausslegendre`.
                Default: `gausslegendre` if no method is provided.
            internal_batch_size: Divides total #steps * #examples
                data points into chunks of size at most internal_batch_size,
                which are computed (forward / backward passes)
                sequentially. internal_batch_size must be at least equal to
                2 * #examples.
                For DataParallel models, each batch is split among the
                available devices, so evaluations on each available
                device contain internal_batch_size / num_devices examples.
                If internal_batch_size is None, then all evaluations are
                processed in one batch.
                Default: None
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output. If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer inputs, otherwise it will be computed with respect
                to layer outputs.
                Note that currently it is assumed that either the input
                or the output of internal layer, depending on whether we
                attribute to the input or output, is a single tensor.
                Support for multiple tensors will be added later.
                Default: False

        Returns:
            Conductance of each neuron in given layer input or
            output. Attributions will always be the same size as
            the input or output of the given layer, depending on
            whether we attribute to the inputs or outputs
            of the layer which is decided by the input flag
            `attribute_to_layer_input`.
            Attributions are returned in a tuple if
            the layer inputs / outputs contain multiple tensors,
            otherwise a single tensor is returned.

        Raises:
            ValueError: if model does not contain conv layers.
            RuntimeError: if attribution has shape (0)
        """

        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)
        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        conductance = self.create_explainer(model=model, layer=layer)
        attributions = conductance.attribute(
            input_data,
            baselines=baselines,
            target=pred_label_idx,
            additional_forward_args=additional_forward_args,
            n_steps=n_steps,
            method=method,
            internal_batch_size=internal_batch_size,
            return_convergence_delta=False,
            attribute_to_layer_input=attribute_to_layer_input,
        )
        validate_result(attributions=attributions)
        return attributions



from typing import List, Tuple

import cv2
import matplotlib
import numpy as np
import torch
import torchvision
from matplotlib.pyplot import Figure

from foxai.array_utils import (
    convert_standardized_float_to_uint8,
    normalize_attributes,
    resize_attributes,
    retain_only_positive,
    standardize_array,
    transpose_array,
)
from foxai.explainer.computer_vision.object_detection.types import ObjectDetectionOutput


def generate_figure(
    attributions: np.ndarray,
    transformed_img: np.ndarray,
    title: str = "",
    figsize: Tuple[int, int] = (8, 8),
    alpha: float = 0.5,
) -> Figure:
    """Create figure from image and heatmap.

    Args:
        attributions: Heatmap.
        transformed_img: Image in shape (H x W x C).
        title: Title of the figure. Defaults to "".
        figsize: Tuple with size of figure. Defaults to (8, 8).
        alpha: Opacity level. Defaults to 0.5,

    Returns:
        Heatmap of single channel applied on original image.
    """
    figure = Figure(figsize=figsize)
    axis = figure.subplots()
    axis.imshow(transformed_img)
    heatmap_plot = axis.imshow(
        attributions, cmap=matplotlib.cm.jet, vmin=0, vmax=1, alpha=alpha
    )

    figure.colorbar(heatmap_plot, label="Pixel relevance")
    axis.get_xaxis().set_visible(False)
    axis.get_yaxis().set_visible(False)
    axis.set_title(title)

    return figure


def _preprocess_img_and_attributes(
    attributes_matrix: np.ndarray,
    transformed_img_np: np.ndarray,
    only_positive_attr: bool,
) -> Tuple[np.ndarray, np.ndarray]:
    """Pre-process image and attributes matrices.

    Pre-processing consists of:
        - squash color dimensions by mean over all colors in attributes matrix
        - optional retension of only positive attributes
        - resizing attributes heatmap to match the size of an image
        - standardization to value range [0-1]
        - transpoze image matrix from (C x H x W) to (H x W x C)

    Args:
        attributions: Features.
        transformed_img: Image in shape (C x H x W) or (H x W).
        only_positive_attr: Whether to display only positive or all attributes.
            Defaults to True.

    Returns:
        Tuple of pre-processed attributes and image matrices.
    """
    single_channel_attributes: np.ndarray = normalize_attributes(
        attributes=attributes_matrix,
    )

    if only_positive_attr:
        single_channel_attributes = retain_only_positive(
            array=single_channel_attributes
        )

    resized_attributes: np.ndarray = resize_attributes(
        attributes=single_channel_attributes,
        dest_height=transformed_img_np.shape[1],
        dest_width=transformed_img_np.shape[2],
    )

    # standardize attributes to uint8 type and back-scale them to range 0-1
    grayscale_attributes = standardize_array(resized_attributes)

    # standardize image
    standardized_img = standardize_array(transformed_img_np.astype(np.dtype(float)))

    # transpoze image from (C x H x W) shape to (H x W x C) to matplotlib imshow
    normalized_transformed_img = transpose_array(
        convert_standardized_float_to_uint8(standardized_img),
    )
    return grayscale_attributes, normalized_transformed_img


def mean_channels_visualization(
    attributions: torch.Tensor,
    transformed_img: torch.Tensor,
    title: str = "",
    figsize: Tuple[int, int] = (8, 8),
    alpha: float = 0.5,
    only_positive_attr: bool = True,
) -> Figure:
    """Create image with calculated heatmap.

    Args:
        attributions: Features.
        transformed_img: Image in shape (C x H x W) or (H x W).
        title: Title of the figure. Defaults to "".
        figsize: Tuple with size of figure. Defaults to (8, 8).
        alpha: Opacity level. Defaults to 0.5,
        only_positive_attr: Whether to display only positive or all attributes.
            Defaults to True.

    Returns:
        Heatmap of mean channel values applied on original image.
    """
    attributes_matrix: np.ndarray = attributions.detach().cpu().numpy()
    transformed_img_np: np.ndarray = transformed_img.detach().cpu().numpy()

    grayscale_attributes, normalized_transformed_img = _preprocess_img_and_attributes(
        attributes_matrix=attributes_matrix,
        transformed_img_np=transformed_img_np,
        only_positive_attr=only_positive_attr,
    )

    return generate_figure(
        attributions=grayscale_attributes,
        transformed_img=normalized_transformed_img,
        title=title,
        figsize=figsize,
        alpha=alpha,
    )


def single_channel_visualization(
    attributions: torch.Tensor,
    transformed_img: torch.Tensor,
    selected_channel: int,
    title: str = "",
    figsize: Tuple[int, int] = (8, 8),
    alpha: float = 0.5,
    only_positive_attr: bool = True,
) -> Figure:
    """Create image with calculated heatmap.

    Args:
        attributions: Features.
        transformed_img: Image in shape (C x H x W) or (H x W).
        selected_channel: Single color channel to visualize.
        title: Title of the figure. Defaults to "".
        figsize: Tuple with size of figure. Defaults to (8, 8).
        alpha: Opacity level. Defaults to 0.5,
        only_positive_attr: Whether to display only positive or all attributes.
            Defaults to True.

    Returns:
        Heatmap of single channel applied on original image.

    Raises:
        ValueError: if selected channel is negative number or exceed dimension
            of color channels of attributes.
    """
    condition: bool = 0 <= selected_channel < attributions.shape[0]
    if not condition:
        raise ValueError(
            f"The selected channel exceeds color dimension. Selected channel: {selected_channel}",
        )

    attributes_matrix: np.ndarray = attributions.detach().cpu().numpy()
    transformed_img_np: np.ndarray = transformed_img.detach().cpu().numpy()
    attributes_matrix = attributes_matrix[selected_channel]

    grayscale_attributes, normalized_transformed_img = _preprocess_img_and_attributes(
        attributes_matrix=attributes_matrix,
        transformed_img_np=transformed_img_np,
        only_positive_attr=only_positive_attr,
    )

    return generate_figure(
        attributions=grayscale_attributes,
        transformed_img=normalized_transformed_img,
        title=title,
        figsize=figsize,
        alpha=alpha,
    )


def preprocess_object_detection_image(input_image: torch.Tensor) -> np.ndarray:
    """Process input image to display.

    Args:
        input_image: Original image of type float in range [0-1].

    Returns:
        Converted image as np.ndarray in (C x H x W).
    """
    return (
        input_image.squeeze(0)
        .mul(255)
        .clamp_(0, 255)
        .permute(1, 2, 0)
        .detach()
        .cpu()
        .numpy()
    )


def get_heatmap_bbox(
    heatmap: np.ndarray,
    bbox: List[int],
    mask_value: int = 0,
) -> np.ndarray:
    """_summary_

    Code based on https://github.com/pooya-mohammadi/deep_utils/blob/main/deep_utils/utils/box_utils/boxes.py.

    Args:
        heatmap: Heatmap to visualize.
        bbox: Bounding box of detection.
        mask_value: Masking value . Defaults to 0.

    Returns:
        Numpy array with heatmap only present in area of given bounding box.
    """
    # fill the outer area of the selected box
    mask = np.ones_like(heatmap, dtype=np.uint8) * mask_value
    mask[bbox[0] : bbox[2], bbox[1] : bbox[3]] = 1
    masked_heatmap = cv2.multiply(heatmap, mask)
    return masked_heatmap


def draw_heatmap_in_bbox(
    bbox: List[int],
    heatmap: torch.Tensor,
    img: np.ndarray,
) -> np.ndarray:
    """Draw heatmap in bounding box on image.

    Args:
        bbox: List of coordinates for bounding box.
        heatmap: Heatmap to display.
        img: Original image.

    Returns:
        Image with displayed heatmap in bounding box area.
    """
    heatmap_np = preprocess_object_detection_image(heatmap).astype(np.uint8)
    heatmap_np = cv2.applyColorMap(heatmap_np, cv2.COLORMAP_JET)

    masked_heatmap = get_heatmap_bbox(heatmap=heatmap_np, bbox=bbox).astype(np.float32)

    img = cv2.add(img, masked_heatmap)
    img = img / img.max()
    img = (img * 255).astype(np.uint8)
    return img


def concat_images(images: List[np.ndarray]) -> np.ndarray:
    """Concatenate images into one.

    Args:
        images: List of images to merge.

    Returns:
        Final image.
    """
    w, h = images[0].shape[:2]
    width = w
    height = h * len(images)
    base_img = np.zeros((width, height, 3), dtype=np.uint8)
    for i, img in enumerate(images):
        base_img[:, h * i : h * (i + 1), ...] = img

    return base_img


def object_detection_visualization(
    detections: ObjectDetectionOutput,
    input_image: torch.Tensor,
) -> np.ndarray:
    """Create array with detection heatmaps.

    Args:
        detections: Object detection data class.
        input_image: Image in shape (C x H x W) or (H x W).

    Returns:
        Array of series of images with heatmap displayed on detection bounding boxes.
    """
    masks = detections.saliency_maps
    boxes = [pred.bbox for pred in detections.predictions]
    class_names = [pred.class_name for pred in detections.predictions]
    img_to_display = preprocess_object_detection_image(input_image)
    img_to_display = img_to_display[..., ::-1]  # convert to bgr
    images = [img_to_display]

    for i, mask in enumerate(masks):
        res_img = img_to_display.copy()
        bbox, cls_name = boxes[i], class_names[i]
        bbox = [int(val) for val in bbox]
        res_img = draw_heatmap_in_bbox(bbox, mask, res_img)

        # convert to (C x H x W)
        res_img_tensor = torch.tensor(res_img).transpose(0, 2).transpose(1, 2)
        bbox = [bbox[1], bbox[0], bbox[3], bbox[2]]
        res_img_tensor = torchvision.utils.draw_bounding_boxes(
            image=res_img_tensor,
            boxes=torch.tensor([bbox]),
            labels=[cls_name],
        )
        # convert to (H x W x C)
        res_img = res_img_tensor.transpose(1, 2).transpose(0, 2).numpy()
        images.append(res_img)

    final_image = concat_images(images)
    return final_image



"""File with Gradient SHAP algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/gradient_shap.py
and https://github.com/pytorch/captum/blob/master/captum/attr/_core/layer/layer_gradient_shap.py.
"""

from abc import abstractmethod
from typing import Any, Optional, Tuple, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import GradientShap, LayerGradientShap

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import get_last_conv_model_layer


class BaseGradientSHAPCVExplainer(Explainer):
    """Base Gradient SHAP algorithm explainer."""

    @abstractmethod
    def create_explainer(
        self,
        model: torch.nn.Module,
        multiply_by_inputs: bool = True,
        **kwargs,
    ) -> Union[GradientShap, LayerGradientShap]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.
            multiply_by_inputs: Indicates whether to factor
                model inputs' multiplier in the final attribution scores.
                In the literature this is also known as local vs global
                attribution. If inputs' multiplier isn't factored in
                then this type of attribution method is also called local
                attribution. If it is, then that type of attribution
                method is called global.
                More detailed can be found here:
                https://arxiv.org/abs/1711.06104

                In case of gradient shap, if `multiply_by_inputs`
                is set to True, the sensitivity scores of scaled inputs
                are being multiplied by (inputs - baselines).

        Returns:
            Explainer object.
        """

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        baselines: Union[None, int, float, torch.Tensor] = None,
        n_samples: int = 5,
        stdevs: Union[float, Tuple[float, ...]] = 0.0,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Gradient SHAP algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which SHAP attribution
                values are computed. If `forward_func` takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            baselines:
                Baselines define the starting point from which expectation
                is computed and can be provided as:

                - a single tensor, if inputs is a single tensor, with
                    the first dimension equal to the number of examples
                    in the baselines' distribution. The remaining dimensions
                    must match with input tensor's dimension starting from
                    the second dimension.

                It is recommended that the number of samples in the baselines'
                tensors is larger than one.
            n_samples: The number of randomly generated examples
                per sample in the input batch. Random examples are
                generated by adding gaussian random noise to each sample.
                Default: `5` if `n_samples` is not provided.
            stdevs: The standard deviation
                of gaussian noise with zero mean that is added to each
                input in the batch. If `stdevs` is a single float value
                then that same value is used for all inputs. If it is
                a tuple, then it must have the same length as the inputs
                tuple. In this case, each stdev value in the stdevs tuple
                corresponds to the input with the same index in the inputs
                tuple.
                Default: 0.0
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It can contain a tuple of ND tensors or
                any arbitrary python type of any shape.
                In case of the ND tensor the first dimension of the
                tensor must correspond to the batch size. It will be
                repeated for each `n_steps` for each randomly generated
                input sample.

                Note that the gradients are not computed with respect
                to these arguments.
                Default: None
            attribute_to_layer_input (bool, optional): Indicates whether to
                compute the attribution with respect to the layer input
                or output. If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer input, otherwise it will be computed with respect
                to layer output.

                Note that currently it is assumed that either the input
                or the output of internal layer, depending on whether we
                attribute to the input or output, is a single tensor.
                Support for multiple tensors will be added later.
                Default: False

        Returns:
            Attribution score computed based on GradientSHAP with respect
            to each input feature. Attributions will always be
            the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)
        gradient_shap = self.create_explainer(model=model, layer=layer)

        # defining baseline distribution of images
        if baselines is None:
            baselines = torch.randn(
                (
                    2 * input_data.shape[0],
                    *input_data.shape[1:],
                ),
                requires_grad=True,
                device=input_data.device,
            )

        if isinstance(gradient_shap, LayerGradientShap):
            attributions = gradient_shap.attribute(
                input_data,
                n_samples=n_samples,
                stdevs=stdevs,
                baselines=baselines,
                target=pred_label_idx,
                return_convergence_delta=False,
                additional_forward_args=additional_forward_args,
                attribute_to_layer_input=attribute_to_layer_input,
            )
        else:
            attributions = gradient_shap.attribute(
                input_data,
                n_samples=n_samples,
                stdevs=stdevs,
                baselines=baselines,
                target=pred_label_idx,
                return_convergence_delta=False,
                additional_forward_args=additional_forward_args,
            )
        validate_result(attributions=attributions)
        return attributions


class GradientSHAPCVExplainer(BaseGradientSHAPCVExplainer):
    """Gradient SHAP algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        multiply_by_inputs: bool = True,
        **kwargs,
    ) -> Union[GradientShap, LayerGradientShap]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.
            multiply_by_inputs: Indicates whether to factor
                model inputs' multiplier in the final attribution scores.
                In the literature this is also known as local vs global
                attribution. If inputs' multiplier isn't factored in
                then this type of attribution method is also called local
                attribution. If it is, then that type of attribution
                method is called global.
                More detailed can be found here:
                https://arxiv.org/abs/1711.06104

                In case of gradient shap, if `multiply_by_inputs`
                is set to True, the sensitivity scores of scaled inputs
                are being multiplied by (inputs - baselines).

        Returns:
            Explainer object.
        """

        return GradientShap(
            forward_func=model,
            multiply_by_inputs=multiply_by_inputs,
        )


class LayerGradientSHAPCVExplainer(BaseGradientSHAPCVExplainer):
    """Layer Gradient SHAP algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        multiply_by_inputs: bool = True,
        layer: Optional[torch.nn.Module] = None,
        **kwargs,
    ) -> Union[GradientShap, LayerGradientShap]:
        """Create explainer object.

        Uses parameter `layer` from `kwargs`. If not provided function will call
        `get_last_conv_model_layer` function to obtain last `torch.nn.Conv2d` layer
        from provided model.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.
                Default: None
            multiply_by_inputs: Indicates whether to factor
                model inputs' multiplier in the final attribution scores.
                In the literature this is also known as local vs global
                attribution. If inputs' multiplier isn't factored in
                then this type of attribution method is also called local
                attribution. If it is, then that type of attribution
                method is called global.
                More detailed can be found here:
                https://arxiv.org/abs/1711.06104

                In case of layer gradient shap, if `multiply_by_inputs`
                is set to True, the sensitivity scores for scaled inputs
                are being multiplied by
                layer activations for inputs - layer activations for baselines.

        Returns:
            Explainer object.

        Raises:
            ValueError: if model does not contain conv layers.
        """

        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        return LayerGradientShap(
            forward_func=model,
            layer=layer,
            multiply_by_inputs=multiply_by_inputs,
        )



"""File with GradCAM algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/guided_grad_cam.py
and https://github.com/pytorch/captum/blob/master/captum/attr/_core/layer/grad_cam.py.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn.functional as F
from captum._utils.typing import TargetType
from captum.attr import GuidedGradCam

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import (
    get_last_conv_model_layer,
    modify_modules,
)
from foxai.explainer.computer_vision.object_detection.base_object_detector import (
    BaseObjectDetector,
)
from foxai.explainer.computer_vision.object_detection.types import ObjectDetectionOutput


class LayerBaseGradCAM(ABC):
    """Layer GradCAM for object detection task."""

    def __init__(
        self,
        target_layer: torch.nn.Module,
    ):
        self._gradients: Dict[str, torch.Tensor] = {}
        self._activations: Dict[str, torch.Tensor] = {}
        self.target_layer = target_layer

        def backward_hook(
            module,  # pylint: disable = (unused-argument)
            grad_input,  # pylint: disable = (unused-argument)
            grad_output,
        ):
            self._gradients["value"] = grad_output[0]

        def forward_hook(
            module,  # pylint: disable = (unused-argument)
            input,  # pylint: disable = (unused-argument,redefined-builtin)
            output,
        ):
            self._activations["value"] = output

        self.target_layer.register_forward_hook(forward_hook)
        self.target_layer.register_backward_hook(backward_hook)

    @property
    def activations(self) -> torch.Tensor:
        return self._activations["value"]

    @property
    def gradients(self) -> torch.Tensor:
        return self._gradients["value"]

    @abstractmethod
    def forward(
        self,
        input_img: torch.Tensor,
    ) -> Union[torch.Tensor, ObjectDetectionOutput]:
        """Forward pass of GradCAM aglorithm.

        Args:
            input_img: Input image with shape of (B, C, H, W).

        Returns:
            ObjectDetectionOutput object for object detection and tensor with saliency
            map for classification.
        """

    def get_saliency_map(
        self,
        height: int,
        width: int,
        gradients: torch.Tensor,
        activations: torch.Tensor,
    ) -> torch.Tensor:
        """Generate saliency map.

        Args:
            height: Original image height.
            width: Original image width.
            gradients: Layer gradients.
            activations: Layer activations.

        Returns:
            Saliency map.
        """
        b, k, _, _ = gradients.size()
        alpha = gradients.view(b, k, -1).mean(2)
        weights = alpha.view(b, k, 1, 1)
        saliency_map = (weights * activations).sum(1, keepdim=True)
        saliency_map = F.relu(saliency_map)
        saliency_map = F.upsample(
            saliency_map, size=(height, width), mode="bilinear", align_corners=False
        )
        saliency_map_min, saliency_map_max = saliency_map.min(), saliency_map.max()
        saliency_map = (
            (saliency_map - saliency_map_min)
            .div(saliency_map_max - saliency_map_min)
            .data
        )
        return saliency_map

    def __call__(
        self,
        input_img: torch.Tensor,
    ) -> Union[torch.Tensor, ObjectDetectionOutput]:
        return self.forward(input_img)


class LayerGradCAMObjectDetection(LayerBaseGradCAM):
    """Layer GradCAM for object detection task."""

    def __init__(
        self,
        model: torch.nn.Module,
        target_layer: torch.nn.Module,
    ):
        super().__init__(target_layer=target_layer)
        self.model = model

    def forward(
        self,
        input_img: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of GradCAM aglorithm.

        Args:
            input_img: Input image with shape of (B, C, H, W).

        Returns:
            Tensor with saliency map.
        """
        saliency_maps: List[torch.Tensor] = []
        _, _, height, width = input_img.size()

        result_list = self.model.forward(input_img)
        for result in result_list:
            score = result.max()

            # clear gradients
            self.model.zero_grad()

            # calculate gradients
            score.backward(retain_graph=True)

            saliency_maps.append(
                self.get_saliency_map(
                    height=height,
                    width=width,
                    gradients=self.gradients,
                    activations=self.activations,
                )
            )
        return torch.cat(saliency_maps)


class BaseGradCAMCVExplainer(Explainer):
    """Base GradCAM algorithm explainer."""

    @abstractmethod
    def create_explainer(
        self, model: torch.nn.Module, layer: torch.nn.Module, **kwargs
    ) -> Union[GuidedGradCam, LayerBaseGradCAM]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.

        Returns:
            Explainer object.
        """

    @abstractmethod
    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate features image with GradCAM algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which attributions
                are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.
                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output in `LayerGradCam`.
                If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer inputs, otherwise it will be computed with respect
                to layer outputs.
                Note that currently it is assumed that either the input
                or the output of internal layer, depending on whether we
                attribute to the input or output, is a single tensor.
                Support for multiple tensors will be added later.
                Default: False

        Returns:
            Element-wise product of (upsampled) GradCAM
            and/or Guided Backprop attributions.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.
            Attributions will be the same size as the provided inputs,
            with each value providing the attribution of the
            corresponding input index.
            If the GradCAM attributions cannot be upsampled to the shape
            of a given input tensor, None is returned in the corresponding
            index position.
        """


class GuidedGradCAMCVExplainer(BaseGradCAMCVExplainer):
    """GuidedGradCAM algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: torch.nn.Module,
        **kwargs,
    ) -> Union[GuidedGradCam, LayerBaseGradCAM]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.

        Returns:
            Explainer object.
        """
        model = modify_modules(model)

        return GuidedGradCam(model=model, layer=layer)

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        interpolate_mode: str = "nearest",
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with GradCAM algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which attributions
                are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.
                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            interpolate_mode: Method for interpolation, which
                must be a valid input interpolation mode for
                torch.nn.functional. These methods are
                "nearest", "area", "linear" (3D-only), "bilinear"
                (4D-only), "bicubic" (4D-only), "trilinear" (5D-only)
                based on the number of dimensions of the chosen layer
                output (which must also match the number of
                dimensions for the input tensor). Note that
                the original GradCAM paper uses "bilinear"
                interpolation, but we default to "nearest" for
                applicability to any of 3D, 4D or 5D tensors.
                Default: "nearest"
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output in `LayerGradCam`.
                If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer inputs, otherwise it will be computed with respect
                to layer outputs.
                Note that currently it is assumed that either the input
                or the output of internal layer, depending on whether we
                attribute to the input or output, is a single tensor.
                Support for multiple tensors will be added later.
                Default: False

        Returns:
            Element-wise product of (upsampled) GradCAM
            and/or Guided Backprop attributions.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.
            Attributions will be the same size as the provided inputs,
            with each value providing the attribution of the
            corresponding input index.
            If the GradCAM attributions cannot be upsampled to the shape
            of a given input tensor, None is returned in the corresponding
            index position.

        Raises:
            ValueError: if model does not contain conv layers.
            RuntimeError: if attributions has shape (0)
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)
        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        guided_cam = self.create_explainer(model=model, layer=layer)

        if isinstance(guided_cam, GuidedGradCam):
            attributions = guided_cam.attribute(
                input_data,
                target=pred_label_idx,
                additional_forward_args=additional_forward_args,
                interpolate_mode=interpolate_mode,
                attribute_to_layer_input=attribute_to_layer_input,
            )
        validate_result(attributions=attributions)
        return attributions


class LayerGradCAMCVExplainer(BaseGradCAMCVExplainer):
    """Layer GradCAM algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: torch.nn.Module,
        **kwargs,
    ) -> Union[GuidedGradCam, LayerBaseGradCAM]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.

        Returns:
            Explainer object.
        """
        model = modify_modules(model)

        return LayerGradCAMObjectDetection(model=model, target_layer=layer)

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,  # pylint: disable = (unused-argument)
        additional_forward_args: Any = None,  # pylint: disable = (unused-argument)
        attribute_to_layer_input: bool = False,  # pylint: disable = (unused-argument)
        relu_attributions: bool = False,  # pylint: disable = (unused-argument)
        **kwargs,
    ) -> torch.Tensor:
        """Generate features image with GradCAM algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            inputs_data: Input for which attributions
                are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
                If forward_func takes multiple tensors as input, a tuple
                of the input tensors should be provided. It is assumed
                that for all given input tensors, dimension 0 corresponds
                to the number of examples, and if multiple input tensors
                are provided, the examples must be aligned appropriately.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.
                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output in `LayerGradCam`.
                If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer inputs, otherwise it will be computed with respect
                to layer outputs.
                Note that currently it is assumed that either the input
                or the output of internal layer, depending on whether we
                attribute to the input or output, is a single tensor.
                Support for multiple tensors will be added later.
                Default: False
            relu_attributions: Indicates whether to
                apply a ReLU operation on the final attribution,
                returning only non-negative attributions. Setting this
                flag to True matches the original GradCAM algorithm,
                otherwise, by default, both positive and negative
                attributions are returned.
                Default: False

        Returns:
            Element-wise product of (upsampled) GradCAM
            and/or Guided Backprop attributions.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.
            Attributions will be the same size as the provided inputs,
            with each value providing the attribution of the
            corresponding input index.
            If the GradCAM attributions cannot be upsampled to the shape
            of a given input tensor, None is returned in the corresponding
            index position.

        Raises:
            ValueError: if model does not contain conv layers.
            RuntimeError: if attributions has shape (0)
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)
        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        gradcam = self.create_explainer(model=model, layer=layer)
        if isinstance(gradcam, LayerBaseGradCAM):
            attributions = gradcam(input_data)

        if not isinstance(attributions, torch.Tensor):
            raise RuntimeError(
                f"Saliency map is `{type(attributions)}`, but expected type is `torch.Tensor`."
            )

        validate_result(attributions=attributions)
        return attributions


class LayerGradCAMObjectDetectionExplainer(LayerBaseGradCAM):
    """Layer GradCAM for object detection task.

    Code based on https://github.com/pooya-mohammadi/yolov5-gradcam.
    """

    def __init__(
        self,
        model: BaseObjectDetector,
        target_layer: torch.nn.Module,
    ):
        super().__init__(target_layer=target_layer)
        self.model = model

    def forward(
        self,
        input_img: torch.Tensor,
    ) -> ObjectDetectionOutput:
        """Forward pass of GradCAM aglorithm.

        Args:
            input_img: Input image with shape of (B, C, H, W).

        Returns:
            ObjectDetectionOutput object.
        """
        saliency_maps: List[torch.Tensor] = []
        _, _, height, width = input_img.size()
        predictions, logits = self.model.forward(input_img)
        for logit, cls in zip(logits[0], [p.class_number for p in predictions]):
            score = logit[cls]
            # clear gradients
            self.model.zero_grad()

            # calculate gradients
            score.backward(retain_graph=True)

            saliency_maps.append(
                self.get_saliency_map(
                    height=height,
                    width=width,
                    gradients=self.gradients,
                    activations=self.activations,
                )
            )
        return ObjectDetectionOutput(
            saliency_maps=saliency_maps,
            logits=logits,
            predictions=predictions,
        )



"""File contains data class for object detection unified model output."""
from dataclasses import dataclass
from typing import List, Tuple

import torch

DetectionOutput = Tuple[torch.Tensor, torch.Tensor]


@dataclass
class PredictionOutput:
    """Data class for model prediction output in YOLO style."""

    bbox: List[int]
    class_number: int
    class_name: str
    confidence: float


@dataclass
class ObjectDetectionOutput:
    """Data class for model predictions for object detection.

    It contains heatmaps, logits and predictions in YOLO style.
    """

    saliency_maps: List[torch.Tensor]
    logits: List[torch.Tensor]
    predictions: List[PredictionOutput]



"""File with Input X Gradient algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/input_x_gradient.py
and https://github.com/pytorch/captum/blob/master/captum/attr/_core/layer/layer_gradient_x_activation.py.
"""

from abc import abstractmethod
from typing import Any, Optional, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import InputXGradient, LayerGradientXActivation

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer
from foxai.explainer.computer_vision.model_utils import get_last_conv_model_layer


class BaseInputXGradientSHAPCVExplainer(Explainer):
    """Base Input X Gradient algorithm explainer."""

    @abstractmethod
    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> Union[InputXGradient, LayerGradientXActivation]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        additional_forward_args: Any = None,
        attribute_to_layer_input: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Input X Gradient algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which
                attributions are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which gradients are computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a tuple
                containing multiple additional arguments including tensors
                or any arbitrary python types. These arguments are provided to
                forward_func in order following the arguments in inputs.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            attribute_to_layer_input: Indicates whether to
                compute the attribution with respect to the layer input
                or output. If `attribute_to_layer_input` is set to True
                then the attributions will be computed with respect to
                layer input, otherwise it will be computed with respect
                to layer output.
                Default: False

        Returns:
            The input x gradient with respect to each input feature or gradient
            and activation for each neuron in given layer output. Attributions
            will always be the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        layer: Optional[torch.nn.Module] = kwargs.get("layer", None)
        input_x_gradient = self.create_explainer(model=model, layer=layer)

        if isinstance(input_x_gradient, LayerGradientXActivation):
            attributions = input_x_gradient.attribute(
                input_data,
                target=pred_label_idx,
                additional_forward_args=additional_forward_args,
                attribute_to_layer_input=attribute_to_layer_input,
            )
        else:
            attributions = input_x_gradient.attribute(
                input_data,
                target=pred_label_idx,
                additional_forward_args=additional_forward_args,
            )
        validate_result(attributions=attributions)
        return attributions


class InputXGradientCVExplainer(BaseInputXGradientSHAPCVExplainer):
    """Input X Gradient algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        **kwargs,
    ) -> Union[InputXGradient, LayerGradientXActivation]:
        """Create explainer object.

        Args:
            model: The forward function of the model or any
                modification of it.

        Returns:
            Explainer object.
        """

        return InputXGradient(
            forward_func=model,
        )


class LayerInputXGradientCVExplainer(BaseInputXGradientSHAPCVExplainer):
    """Layer Input X Gradient algorithm explainer."""

    def create_explainer(
        self,
        model: torch.nn.Module,
        layer: Optional[torch.nn.Module] = None,
        multiply_by_inputs: bool = True,
        **kwargs,
    ) -> Union[InputXGradient, LayerGradientXActivation]:
        """Create explainer object.

        Uses parameter `layer` from `kwargs`. If not provided function will call
        `get_last_conv_model_layer` function to obtain last `torch.nn.Conv2d` layer
        from provided model.

        Args:
            model: The forward function of the model or any
                modification of it.
            layer: Layer for which attributions are computed.
                Output size of attribute matches this layer's input or
                output dimensions, depending on whether we attribute to
                the inputs or outputs of the layer, corresponding to
                attribution of each neuron in the input or output of
                this layer.
                Default: None
            multiply_by_inputs: Indicates whether to factor
                model inputs' multiplier in the final attribution scores.
                In the literature this is also known as local vs global
                attribution. If inputs' multiplier isn't factored in,
                then this type of attribution method is also called local
                attribution. If it is, then that type of attribution
                method is called global.
                More detailed can be found here:
                https://arxiv.org/abs/1711.06104

                In case of layer gradient x activation, if `multiply_by_inputs`
                is set to True, final sensitivity scores are being multiplied by
                layer activations for inputs.

        Returns:
            Explainer object.

        Raises:
            ValueError: if model does not contain conv layers.
        """

        if layer is None:
            layer = get_last_conv_model_layer(model=model)

        return LayerGradientXActivation(
            forward_func=model,
            layer=layer,
            multiply_by_inputs=multiply_by_inputs,
        )



"""File with Occulusion algorithm explainer classes.

Based on https://github.com/pytorch/captum/blob/master/captum/attr/_core/occlusion.py.
"""

from typing import Any, Tuple, Union

import torch
from captum._utils.typing import TargetType
from captum.attr import Occlusion

from foxai.array_utils import validate_result
from foxai.explainer.base_explainer import Explainer


class OcclusionCVExplainer(Explainer):
    """Occlusion algorithm explainer."""

    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,
        pred_label_idx: TargetType = None,
        sliding_window_shapes: Union[Tuple[int, ...], Tuple[Tuple[int, ...], ...]] = (
            1,
            1,
            1,
        ),
        strides: Union[
            None, int, Tuple[int, ...], Tuple[Union[int, Tuple[int, ...]], ...]
        ] = None,
        baselines: Union[None, int, float, torch.Tensor] = None,
        additional_forward_args: Any = None,
        perturbations_per_eval: int = 1,
        show_progress: bool = False,
        **kwargs,
    ) -> torch.Tensor:
        """Generate model's attributes with Occlusion algorithm explainer.

        Args:
            model: The forward function of the model or any
                modification of it.
            input_data: Input for which occlusion
                attributions are computed. If forward_func takes a single
                tensor as input, a single input tensor should be provided.
            pred_label_idx: Output indices for
                which difference is computed (for classification cases,
                this is usually the target class).
                If the network returns a scalar value per example,
                no target index is necessary.
                For general 2D outputs, targets can be either:

                - a single integer or a tensor containing a single
                    integer, which is applied to all input examples

                - a list of integers or a 1D tensor, with length matching
                    the number of examples in inputs (dim 0). Each integer
                    is applied as the target for the corresponding example.

                For outputs with > 2 dimensions, targets can be either:

                - A single tuple, which contains #output_dims - 1
                    elements. This target index is applied to all examples.

                - A list of tuples with length equal to the number of
                    examples in inputs (dim 0), and each tuple containing
                    #output_dims - 1 elements. Each tuple is applied as the
                    target for the corresponding example.

                Default: None
            sliding_window_shapes: Shape of patch
                (hyperrectangle) to occlude each input. For a single
                input tensor, this must be a tuple of length equal to the
                number of dimensions of the input tensor - 1, defining
                the dimensions of the patch. If the input tensor is 1-d,
                this should be an empty tuple.
                Default: (1, 1, 1)
            strides:
                This defines the step by which the occlusion hyperrectangle
                should be shifted by in each direction for each iteration.
                For a single tensor input, this can be either a single
                integer, which is used as the step size in each direction,
                or a tuple of integers matching the number of dimensions
                in the occlusion shape, defining the step size in the
                corresponding dimension.
                To ensure that all inputs are covered by at least one
                sliding window, the stride for any dimension must be
                <= the corresponding sliding window dimension if the
                sliding window dimension is less than the input
                dimension.
                If None is provided, a stride of 1 is used for each
                dimension of each input tensor.
                Default: None
            baselines:
                Baselines define reference value which replaces each
                feature when occluded.
                Baselines can be provided as:

                - a single tensor, if inputs is a single tensor, with
                    exactly the same dimensions as inputs or
                    broadcastable to match the dimensions of inputs

                - a single scalar, if inputs is a single tensor, which will
                    be broadcasted for each input value in input tensor.

                In the cases when `baselines` is not provided, we internally
                use zero scalar corresponding to each input tensor.
                Default: None
            additional_forward_args: If the forward function
                requires additional arguments other than the inputs for
                which attributions should not be computed, this argument
                can be provided. It must be either a single additional
                argument of a Tensor or arbitrary (non-tuple) type or a
                tuple containing multiple additional arguments including
                tensors or any arbitrary python types. These arguments
                are provided to forward_func in order following the
                arguments in inputs.
                For a tensor, the first dimension of the tensor must
                correspond to the number of examples. For all other types,
                the given argument is used for all forward evaluations.
                Note that attributions are not computed with respect
                to these arguments.
                Default: None
            perturbations_per_eval: Allows multiple occlusions
                to be included in one batch (one call to forward_fn).
                By default, perturbations_per_eval is 1, so each occlusion
                is processed individually.
                Each forward pass will contain a maximum of
                perturbations_per_eval * #examples samples.
                For DataParallel models, each batch is split among the
                available devices, so evaluations on each available
                device contain at most
                (perturbations_per_eval * #examples) / num_devices
                samples.
                Default: 1
            show_progress: Displays the progress of computation.
                It will try to use tqdm if available for advanced features
                (e.g. time estimation). Otherwise, it will fallback to
                a simple output of progress.
                Default: False

        Returns:
            The attributions with respect to each input feature.
            Attributions will always be
            the same size as the provided inputs, with each value
            providing the attribution of the corresponding input index.
            If a single tensor is provided as inputs, a single tensor is
            returned. If a tuple is provided for inputs, a tuple of
            corresponding sized tensors is returned.

        Raises:
            RuntimeError: if attribution has shape (0).
        """
        occlusion = Occlusion(model)

        # defining baseline distribution of images
        if baselines is None:
            baselines = torch.randn(
                input_data.shape,
                requires_grad=True,
                device=input_data.device,
            )

        attributions = occlusion.attribute(
            input_data,
            strides=strides,
            target=pred_label_idx,
            sliding_window_shapes=sliding_window_shapes,
            baselines=baselines,
            additional_forward_args=additional_forward_args,
            perturbations_per_eval=perturbations_per_eval,
            show_progress=show_progress,
        )
        validate_result(attributions=attributions)
        return attributions



"""Abstract Explainer class."""
from abc import ABC, abstractmethod
from typing import TypeVar

import torch
from captum._utils.typing import TargetType


class Explainer(ABC):
    """Abstract explainer class."""

    # TODO: add support in explainer for multiple input models
    @abstractmethod
    def calculate_features(
        self,
        model: torch.nn.Module,
        input_data: torch.Tensor,  # TODO: add more generic way of passing model inputs # pylint: disable = (fixme)
        pred_label_idx: TargetType = None,
        **kwargs,
    ) -> torch.Tensor:
        """Calculate features of given explainer.

        Args:
            model: Neural network model You want to explain.
            input_data: Input image.
            pred_label_idx: Predicted label.

        Returns:
            Tensor of attributes.
        """

    @property
    def algorithm_name(self) -> str:
        """Get algorithm name.

        Returns:
            str: Name of algorithm.
        """
        return type(self).__name__


CVExplainerT = TypeVar("CVExplainerT", bound=Explainer)
"""CVExplainer subclass type."""



"""File contains abstract base ObjectDetector class."""
from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import torch
from torch import nn

from foxai.explainer.computer_vision.object_detection.types import PredictionOutput
from foxai.explainer.computer_vision.object_detection.utils import resize_image


class BaseObjectDetector(nn.Module, ABC):
    """Base ObjectDetector class which returns predictions with logits to explain.

    Code based on https://github.com/pooya-mohammadi/yolov5-gradcam.
    """

    @abstractmethod
    def forward(
        self,
        image: torch.Tensor,
    ) -> Tuple[List[PredictionOutput], List[torch.Tensor]]:
        """Forward pass of the network.

        Args:
            image: Image to process.

        Returns:
            Tuple of 2 values, first is tuple of predictions containing bounding-boxes,
            class number, class name and confidence; second value is list of tensors
            with logits per each detection.
        """

    @staticmethod
    def preprocessing(
        img: np.ndarray,
        new_shape: Tuple[int, int] = (640, 640),
        change_original_ratio: bool = False,
        scaleup: bool = True,
    ) -> torch.Tensor:
        """Preprocess image before prediction.

        Preprocessing is a process consisting of steps:
        * adding batch dimension
        * resizing images to desired shapes
        * adjusting image channels to (B x C x H x W)
        * convertion to float

        Args:
            img: Image to preprocess.
            new_shape: Desired shape of image. Defaults to (640, 640).
            change_original_ratio: If resized image should have different height to
                width ratio than original image. Defaults to False.
            scaleup: If scale up image. Defaults to True.

        Returns:
            Tensor containing preprocessed image.
        """
        if len(img.shape) != 4:
            # add batch dimension
            img = np.expand_dims(img, axis=0)

        # resize all images from batch
        img = np.array(
            [
                resize_image(
                    image=im,
                    new_shape=new_shape,
                    change_original_ratio=change_original_ratio,
                    scaleup=scaleup,
                )
                for im in img
            ]
        )
        # convert array from (B x H x W x C) to (B x C x H x W)
        img = img.transpose((0, 3, 1, 2))
        img_tensor = torch.from_numpy(img)

        # convert from uint8 to float
        img_tensor = img_tensor / 255.0
        return img_tensor



"""File contains functions to handle numpy arrays."""
import sys

import cv2
import numpy as np
import torch


def standardize_array(array: np.ndarray) -> np.ndarray:
    """Standardize array values to range [0-1].

    Args:
        array: Numpy array of floats.

    Returns:
        Numpy array with scaled values.

    Raises:
        ValueError: if array is not of type np.float.
    """
    if not array.dtype == np.dtype(float):
        raise ValueError(
            f"Array should be of type: np.float, current type: {array.dtype}"
        )

    return (array - np.min(array)) / (
        (np.max(array) - np.min(array)) + sys.float_info.epsilon
    )


def convert_standardized_float_to_uint8(array: np.ndarray) -> np.ndarray:
    """Convert float standardize float array to uint8 with values scaling.

    Args:
        array: Numpy array of floats.

    Returns:
        Numpy array with scaled values with type np.uint8.

    Raises:
        ValueError: if array is not of type np.float.
    """
    if not array.dtype == np.dtype(float):
        raise ValueError(
            f"Array should be of type: np.float, current type: {array.dtype}"
        )

    return (array * 255).astype(np.uint8)


def retain_only_positive(array: np.ndarray) -> np.ndarray:
    """Retain only positive values from array.

    Args:
        array: Array.

    Returns:
        Array with negative values replaced by zero.
    """
    array[array < 0] = 0
    return array


def normalize_attributes(
    attributes: np.ndarray,
) -> np.ndarray:
    """Normalize attributes.

    For attributes with color dimension calculate mean over all colors.

    Args:
        attributes: Array of attributes.

    Returns:
        Single channel array of attributes.

    Raises:
        ValueError: if shape of attribute array is incorrect.
    """
    ret_array: np.ndarray
    if len(attributes.shape) == 3:
        # if we have attributes with shape (C x H x W)
        # where C is color, W is width and H is height dimension
        # calculate mean over attributes for all colors
        ret_array = np.mean(attributes, axis=0)
    elif len(attributes.shape) == 2:
        ret_array = attributes
    else:
        raise ValueError(f"Incorrect shape of attributions: {attributes.shape}")

    return ret_array


def resize_attributes(
    attributes: np.ndarray,
    dest_width: int,
    dest_height: int,
) -> np.ndarray:
    """Resize attributes to match desired shape.

    Args:
        attributes: Array of attributes.
        dest_width: Desired width of attributes array.
        dest_height: Desired height of attributes array.

    Returns:
        Resized attributes array.
    """
    # resize attributes matrix to match input image
    single_channel_attributes: np.ndarray = np.array(
        cv2.resize(
            attributes,
            (dest_width, dest_height),
        ),
        dtype=np.dtype(float),
    )

    return single_channel_attributes


def transpose_array(array: np.ndarray) -> np.ndarray:
    """Transpoze array from (C x H x W) to (H x W x C) shape.

    C stands for color, H stands for height and W stands for width.

    Args:
        array: Array of shape (C x H x W).

    Returns:
        Array of shape (H x W x C).
    """
    return np.transpose(array, (1, 2, 0))


def validate_result(attributions: torch.Tensor) -> None:
    """Validate calculated attributes.

    Args:
        attributions: Tensor with calculated attributions.

    Raises:
        RuntimeError if tensor is empty.
    """
    if attributions.shape[0] == 0:
        raise RuntimeError(
            "Error occured during attribution calculation. "
            + "Make sure You are applying this method to CNN network.",
        )



# pylint: disable = missing-module-docstring
import logging
from logging import StreamHandler

__LOG_FORMAT: str = "%(asctime)s %(levelname)s %(name)s - %(message)s"


def create_logger(logger_name: str, level: int = logging.DEBUG) -> logging.Logger:
    """Create logger with the goven name.

    Args:
        logger_name: the name of the logger

    Return:
        created logger object
    """

    result: logging.Logger = logging.getLogger(logger_name)
    result.setLevel(level)
    result.propagate = False

    formatter: logging.Formatter = logging.Formatter(__LOG_FORMAT)
    console_handler = StreamHandler()
    console_handler.setFormatter(formatter)
    result.addHandler(console_handler)
    result.propagate = True

    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['foxai',
 'foxai.callbacks',
 'foxai.cli',
 'foxai.explainer',
 'foxai.explainer.computer_vision',
 'foxai.explainer.computer_vision.algorithm',
 'foxai.explainer.computer_vision.object_detection']

package_data = \
{'': ['*']}

install_requires = \
['captum>=0.5.0,<1.0.0',
 'hydra-core>=1.3.1,<2.0.0',
 'numpy>=1.21.4,<2.0.0',
 'opencv-python>=4.7.0.68,<5.0.0',
 'pandas>=1.1.0,<2.0.0',
 'pytorch-lightning>=1.5.0,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'single-source>=0.3.0,<1.0.0',
 'torch>=1.12.1,<2.0.0',
 'wandb>=0.13.7,<1.0.0']

entry_points = \
{'console_scripts': ['foxai-wandb-updater = foxai.cli.experiment_update:main']}

setup_kwargs = {
    'name': 'foxai',
    'version': '0.6.0',
    'description': 'Model Interpretability for PyTorch.',
    'long_description': '# FoXAI\n\nFoXAI simplifies the application of e**X**plainable **AI** algorithms to explain the\nperformance of neural network models during training. The library acts as an\naggregator of existing libraries with implementations of various XAI algorithms and\nseeks to facilitate and popularize their use in machine learning projects.\n\nCurrently, only algorithms related to computer vision are supported, but we plan to\nadd support for text, tabular and multimodal data problems in the future.\n\n## Table of content:\n* [Installation](#installation)\n    * [GPU acceleration](#gpu-acceleration)\n    * [Manual installation](#manual-installation)\n* [Getting started](#getting-started)\n* [Development](#development)\n    * [Requirements](#requirements)\n    * [CUDA](#cuda)\n    * [pyenv](#pyenv)\n    * [Poetry](#poetry)\n    * [pre-commit hooks](#pre-commit-hooks-setup)\n    * [Note](#note)\n        * [Artifacts directory structure](#artifacts-directory-structure)\n        * [Examples](#examples)\n\n# Installation\n\nInstallation requirements:\n* `Python` >=3.7.2,<3.11\n\n**Important**: For any problems regarding installation we advise to refer first to our [FAQ](FAQ.md).\n\n## GPU acceleration\n\nTo use the torch library with GPU acceleration, you need to install\na dedicated version of torch with support for the installed version of CUDA\ndrivers in the version supported by the library, at the moment `torch>=1.12.1,<2.0.0`.\nA list of `torch` wheels with CUDA support can be found at\n[https://download.pytorch.org/whl/torch/](https://download.pytorch.org/whl/torch/).\n\n## Manual installation\n\nIf you would like to install from the source you can build a `wheel` package using `poetry`.\nThe assumption is that the `poetry` package is installed. You can find how to install\n`poetry` [here](#poetry). To build `wheel` package run:\n\n```bash\ngit clone https://github.com/softwaremill/FoXAI.git\ncd FoXAI/\npoetry install\npoetry build\n```\n\nAs a result you will get `wheel` file inside `dist/` directory that you can install\nvia `pip`:\n```bash\npip install dist/foxai-x.y.z-py3-none-any.whl\n```\n\n# Getting started\n\nTo use the FoXAI library in your ML project, simply add an additional object of type\n`WandBCallback` to the `Trainer`\'s callback list from the `pytorch-lightning` library.\nCurrently, only the Weights and Biases tool for tracking experiments is supported.\n\nBelow is a code snippet from the example (`example/mnist_wandb.py`):\n\n```python\nimport torch\nfrom pytorch_lightning import Trainer\nfrom pytorch_lightning.loggers import WandbLogger\n\nimport wandb\nfrom foxai.callbacks.wandb_callback import WandBCallback\nfrom foxai.context_manager import CVClassificationExplainers, ExplainerWithParams\n\n    ...\n    wandb.login()\n    wandb_logger = WandbLogger(project=project_name, log_model="all")\n    callback = WandBCallback(\n        wandb_logger=wandb_logger,\n        explainers=[\n            ExplainerWithParams(\n                explainer_name=CVClassificationExplainers.CV_INTEGRATED_GRADIENTS_EXPLAINER\n            ),\n            ExplainerWithParams(\n                explainer_name=CVClassificationExplainers.CV_GRADIENT_SHAP_EXPLAINER\n            ),\n        ],\n        idx_to_label={index: index for index in range(0, 10)},\n    )\n    model = LitMNIST()\n    trainer = Trainer(\n        accelerator="gpu",\n        devices=1 if torch.cuda.is_available() else None,\n        max_epochs=max_epochs,\n        logger=wandb_logger,\n        callbacks=[callback],\n    )\n    trainer.fit(model)\n```\n\n## CLI\n\nA CLI tool is available to update the artifacts of an experiment tracked in\nWeights and Biases. Allows you to create XAI explanations and send them to\nW&B offline. This tool is using `hydra` to handle the configuration of `yaml` files.\nTo check options type:\n\n```bash\nfoxai-wandb-updater --help\n```\n\nTypical usage with configuration in `config/config.yaml`:\n```bash\nfoxai-wandb-updater --config-dir config/ --config-name config\n```\n\nContent of `config.yaml`:\n```bash\nusername: <WANDB_USERANEM>\nexperiment: <WANDB_EXPERIMENT>\nrun_id: <WANDB_RUN_ID>\nclassifier: # model class to explain\n  _target_: example.streamlit_app.mnist_model.LitMNIST\n  batch_size: 1\n  data_dir: "."\nexplainers: # list of explainers to use\n - explainer_with_params:\n    explainer_name: CV_GRADIENT_SHAP_EXPLAINER\n    kwargs:\n      n_steps: 1000\n```\n\n# Development\n\n## Requirements\n\nThe project was tested using Python version `3.8`.\n\n## CUDA\n\nThe recommended version of CUDA is `10.2` as it is supported since version\n`1.5.0` of `torch`. You can check the compatibility of your CUDA version\nwith the current version of `torch`:\nhttps://pytorch.org/get-started/previous-versions/.\n\nAs our starting Docker image we were using the one provided by Nvidia: ``nvidia/cuda:10.2-devel-ubuntu18.04``.\n\nIf you wish an easy to use docker image we advise to use our ``Dockerfile``.\n\n## pyenv\nOptional step, but probably one of the easiest way to actually get Python version with all the needed aditional tools (e.g. pip).\n\n`pyenv` is a tool used to manage multiple versions of Python. To install\nthis package follow the instructions on the project repository page:\nhttps://github.com/pyenv/pyenv#installation. After installation You can\ninstall desired Python version, e.g. `3.8.16`:\n```bash\npyenv install 3.8.16\n```\n\nThe next step is required to be able to use a desired version of Python with\n`poetry`. To activate a specific version of Python interpreter execute the command:\n```bash\npyenv local 3.8.16 # or `pyenv global 3.8.16`\n```\n\nInside the repository with `poetry` You can select a specific version of Python\ninterpreter with the command:\n```bash\npoetry env use 3.8.16\n```\n\nAfter changing the interpreter version You have to once again install all\ndependencies:\n```bash\npoetry install\n```\n\n## Poetry\n\nTo separate runtime environments for different services and repositories, it is\nrecommended to use a virtual Python environment. You can configure `Poetry` to\ncreate a new virtual environment in the project directory of every repository. To\ninstall `Poetry` follow the instruction at https://python-poetry.org/docs/#installing-with-the-official-installer. We are using `Poetry` in version\n`1.2.1`. To install a specific version You have to provide desired package\nversion:\n```bash\ncurl -sSL https://install.python-poetry.org | POETRY_VERSION=1.2.1 python3 -\n```\nAdd poetry to PATH:\n```bash\nexport PATH="/home/ubuntu/.local/bin:$PATH"\necho \'export PATH="/home/ubuntu/.local/bin:$PATH"\' >> ~/.bashrc\n```\n\nAfter installation, configure the creation of virtual environments in the directory of the project.\n```bash\npoetry config virtualenvs.create true\npoetry config virtualenvs.in-project true\n```\n\nThe final step is to install all the dependencies defined in the\n`pyproject.toml` file.\n\n```bash\npoetry install\n```\n\nOnce all the steps have been completed, the environment is ready to go.\nA virtual environment by default will be created with the name `.venv` inside\nthe project directory.\n\n## Pre-commit hooks setup\n\nTo improve the development experience, please make sure to install\nour [pre-commit](https://pre-commit.com/) hooks as the very first step after\ncloning the repository:\n\n```bash\npoetry run pre-commit install\n```\n\n## Note\n---\nAt the moment only explainable algorithms for image classification are\nimplemented. In the future more algorithms and more computer vision tasks will\nbe introduced. In the end, the module should work with all types of tasks (NLP, etc.).\n\n### Examples\n\nIn `example/notebooks/` directory You can find notebooks with example usage of this\nframework. Scripts in `example/` directory contain samples of training models using\ndifferent callbacks.\n',
    'author': 'ReasonField Lab Team',
    'author_email': 'None',
    'maintainer': 'Adam Jan Kaczmarek',
    'maintainer_email': 'adam.kaczmarek@reasonfieldlab.com',
    'url': 'https://github.com/softwaremill/FoXAI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<3.11',
}


setup(**setup_kwargs)



"""File contains functions to modifiy DNN models."""
from typing import List

import torch


def modify_modules(model: torch.nn.Module) -> torch.nn.Module:
    """Modify modules of given model.

    Function iterates over all modules and sets property `inplace`
    to `False` for every `torch.nn.ReLU` activation function.

    Args:
        model: Neural network object to be modified.

    Returns:
        Modified neural network object.
    """
    for module in model.modules():  # pylint: disable = (duplicate-code)
        if isinstance(module, torch.nn.ReLU):
            module.inplace = False

    return model


def get_last_conv_model_layer(model: torch.nn.Module) -> torch.nn.Module:
    """Get the last convolutional layer from the torch model.

    Args:
        model: torch.nn.Module

    Returns:
       The last convolutional layer of the model.

    Raises:
        ValueError if the model does not contain convolutional layers.
    """

    conv_layers: List[torch.nn.Module] = []
    for module in model.modules():
        if isinstance(module, torch.nn.Conv2d):
            conv_layers.append(module)

    if not conv_layers:
        raise ValueError("The model does not contain convolution layers.")

    return conv_layers[-1]
